"""An implementation of the A* searching algorithm.

dyoo@hkn.eecs.berkeley.edu

I got so disgusted at my previous attempt at A*, so here I go again.
Hopefully this version will be easier on the eyes.

A* is a search algorithm that's similar to Dijkstra's algorithm: given
a graph, a start node, and a goal, A* will search for the shortest
path toward the goal.

To help it get there faster, we can provide a heuristic that evaluates
how far we are from that goal.  With a good heuristic, finding an
optimal solution takes MUCH less time.

The main function that one would use is aStar().  Take a look at
_testAStar() to see how it's used.


jackdied@yahoo.com

Re-jiggered to use generators

"""

from priorityqueue import PriorityQueue
from __future__ import generators

__version__ = "$Revision: 1.3 $"

class AStar:
    def __init__(self, start, goal, neighbor_func, distance_func, heuristic_func):
        self.start = start;
        self.goal = goal;
        self.neighbor_func = neighbor_func
        self.distance_func = distance_func
        self.heuristic_func = heuristic_func

        self.solution = None
        return

    def do_a_little(self):
        pqueue = PriorityQueue()
        g_costs = {self.start : 1}
        parents = {self.start : self.start}

        pqueue.push(self.heuristic_func(self.start, self.goal), self.start)
        while not pqueue.isEmpty():
            next_cost, next_node = pqueue.pop()
            g_costs[next_node] = g_costs[parents[next_node]] \
                                 + self.distance_func(next_node, parents[next_node])
            if next_node == self.goal:
                self.solution = getPathToGoal(self.start, self.goal, parents)
                return
            children = self.neighbor_func(next_node)
            for child in children:
                updateChild(self.goal, self.distance_func, self.heuristic_func,
                            child, next_node, parents, g_costs, pqueue)
                yield None
            yield None
        self.solution = getPathToGoal(self.start, self.goal, parents)
        return
        
def updateChild(goal, distance_func, heuristic_func,
                child, next_node, parents, g_costs, pqueue):
    """Appropriately update the parents, g_costs, and pqueue structures.

    This is a helper function, since aStar() was getting a bit bulky."""
    if g_costs.has_key(child): return
    f = g_costs[next_node] + distance_func(next_node, child) \
        + heuristic_func(child, goal)
    if pqueue.push(f, child):
        parents[child] = next_node

def getPathToGoal(start, goal, parents):
    """Given the hash of parental links, follow back and return the
    chain of ancestors."""
    try:
        results = []
        while goal != start:
            results.append(goal)
            goal = parents[goal]
        results.append(start)
        results.reverse()
        return results
    except KeyError: return []
