# python imports
import re, xreadlines, string
# our imports
import IO

HOST = 'localhost'
PORT = 80

__version__ = "$Revision: 1.10 $"

class Bot:
  """The basic class to store Bot state"""
  def __init__(self, id, carry = 0, money = 0):
    self.id = id
    self.carrymax = carry
    self.carrying = 0
    self.capacity = carry
    self.startmoney = money
    self.money = money
    self.x_pos = None
    self.y_pos = None
    self.curr_loc = None
    self.last_loc = None
    self.packages = {}
    self.package_cnt = 0
    self.deliverme = None
    self.score = 0
    return

  def player_string(self):
    return "%d %d %d\n" % (self.id, self.carrymax, self.startmoney)

  def set_x(self, x):
    self.x_pos = x
    return

  def set_y(self, y):
    self.y_pos = y
    return

  def move(self, dir):
    """ NB, We invert the map N-S so 1,1 is NW!"""
    self.last_loc = (self.x_pos, self.y_pos)
    if (dir == 'E'):
      self.set_x(self.x_pos + 1)
    elif (dir == 'W'):
      self.set_x(self.x_pos - 1)
    elif (dir == 'S'):
      self.set_y(self.y_pos - 1)
    elif (dir == 'N'):
      self.set_y(self.y_pos + 1)
    else:
      raise 'Bad direction passed to Bot::mov'

    self.curr_loc = (self.x_pos, self.y_pos)
    return

  def rev_move(self, oldpos, newpos):
    """return a direction for two tuples of adjacent coords"""
    if (oldpos[0] < newpos[0]):
      return 'E'
    elif (oldpos[0] > newpos[0]):
      return 'W'
    elif (oldpos[1] > newpos[1]):
      return 'S'
    elif (oldpos[1] < newpos[1]):
      return 'N'
    else:
      raise 'Bad tuple'

  def pickup(self, package):
    """A bot picked up a package"""
    self.packages[package.id] = package
    if (package.weight): # if it is known
      self.carrying += package.weight
      self.capacity -= package.weight
    self.package_cnt += 1
    if (not self.deliverme):
      self.deliverme = package
    return

  def drop(self, package):
    if (package.weight): # if it is known
      self.carrying -= package.weight
      self.capacity += package.weight
    del self.packages[package.id]
    self.package_cnt -= 1
    if (self.deliverme == package):
      self.deliverme = None
      if (self.package_cnt):
        self.deliverme = self.packages.values()[0]
        
    return

  def make_move(self, bid, cmd):
    self.money -= bid
    if (self.money <= 0):
      print "Outta cash, gonna die"
    return "%d %s\n" % (bid, cmd)

class Package:
  """The basic class to store package state"""
  def __init__(self, id):
    self.id = id
    self.x_des = None
    self.y_des = None
    self.weight = None
    return

  def set_destination(self, x, y):
    self.x_des = x
    self.y_des = y
    return

  def set_weight(self, weight):
    self.weight = weight
    return

  def myprint(self):
    print 'id %d, weight %d, x_des %d, y_des %d' % (self.id, self.weight, self.x_des, self.y_des)

class Board:
  """The board for the competition, handles state"""

  def __init__(self, width, height):
    """Intializes an empty board of width/height"""
    self.width = width
    self.height = height
    self.bases = {}
    self.maybe_bases = {}
    self.doubtful_bases = {}
    self.full_init()

  def full_init(self):
    width = self.width
    height = self.height
    self._arr = ['BAD'] * (height + 2)
    self._arr[0] = ['#'] * (width + 2)
    self._arr[height+1] = ['#'] * (width + 2)

    self._xarr = ['XTRA'] * (height + 1)
    for (i) in range(1, height + 1):
      self._xarr[i] = ['XTRA'] * (width + 1)
      for (j) in range(1, width + 1):
        self._xarr[i][j] = []

    self._xarr_locs = {}
    
    return

  def set_row(self, row_n, str):
    """parse and store a row string"""
    newrow = [(str[i:i+1]) for (i) in range(len(str))]
    # Edges kill you, mark them as water
    newrow.insert(0,'~')
    newrow.append('~')
    self._arr[row_n] = newrow
    if (len(self._arr[row_n]) != self.width + 2):
      raise 'Parse error, width doesnt match expected'
    return

  def find_bases(self):
    self.bases = {}
    for (i) in range(self.height, 0, -1):
      for (j) in range(1,self.width+1):
        if (self._arr[i][j] == '@'):
          self.bases[(j,i)] = 0
        
    return

  def from_file(self, filename):
    fob = file(filename)
    lines = []
    for (line) in xreadlines.xreadlines(fob):
      lines.insert(0,line.strip())
    fob.close()

    self.width = len(lines[0])
    self.height = len(lines)
    self.full_init()
    for (i) in range(1,self.height+1):
      self.set_row(i, lines[i-1])

    return

  def extra(self, ob, x, y):
    """Add an object to a location"""
    if (self._xarr_locs.has_key(id(ob))):
      (locx, locy) = self._xarr_locs[id(ob)]
      if (x != locx or y != locy): # erase our old info
        self.rm_extra(ob)
      else:
        return# nothing to do
    self._xarr_locs[id(ob)] = [x,y]
    self._xarr[y][x].append(ob)
    return
  
  def extra_here(self, x, y):
    """return the list of stuff here"""
    return self._xarr[y][x]

  def rm_extra(self, ob):
    """Remove an object from a location"""
    obid = id(ob)
    if (not self._xarr_locs.has_key(obid)): # we didn't know about it anyway
      return 0
    (x, y) = self._xarr_locs[obid]
    self._xarr[y][x] = filter(lambda x:id(x) != obid, self._xarr[y][x])
    return len(self._xarr[y][x])

  def len_extra(self, x, y):
    return len(self._xarr[y][x])

  def coords(self, x, y, dir):
    xoff = 0;
    yoff = 0;
    if (dir == 'E'):
      xoff = +1
    elif (dir == 'W'):
      xoff = -1
    elif (dir == 'S'):
      yoff = -1
    elif (dir == 'N'):
      yoff = +1
    else:
      raise 'Bad direction passed to Bot::mov'

    return (x+xoff, y+yoff)

  def peek(self, x, y):
    return self._arr[y][x]

  def myprint(self):
    """pretty printer"""
    str = "%d %d\n" % (self.width, self.height)
    for (i) in range(self.height, 0, -1):
      for (j) in range(1,self.width+1):
        str += self._arr[i][j]
      str += "\n"
    return str

  def mydump(self):
    """pretty printer"""
    str = "%d %d\n" % (self.width, self.height)
    for (i) in range(self.height + 2):
      for (j) in range(self.width + 2):
        str += self._arr[i][j]
      str += "\n"
    return str

  def neighbors(self, x, y):
    answer = []
    for (dir) in ['N','S','E','W']:
      (nx, ny) = self.coords(x,y,dir)
      peek = self.peek(nx,ny)
      if (peek != '~' and peek != '#'):
        answer.append((nx,ny))
    return answer
  
move_re = re.compile('(\w)\s*(.*)')
arg_re = re.compile('(\w)\s+(\d+)\s*(.*)')
id_re = re.compile('(\d+)\s*(.*)')
package_re = re.compile('(-?\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*(.*)')

class World:
  """the One True Singleton"""
  __single = None # our singleton reference
  # the meat of World is in __real_World

  def __init__(self, descriptor = None):
    """New a dummy object that proxys for __real_World,
       this is the instance that all other modules use to interface with World"""
        # Check whether we already have an instance
    if World.__single is None:
      # Create and remember instance
      World.__single = World.__real_World(descriptor)

    # Store instance reference as the only member in the handle
    self.__dict__['_World__single'] = World.__single

    # These are just to fool pychecker, which doesn't know
    # about the singleton trick in Python
    #    self.io = None
    #    self.bot = None # us
    #    self.board = None # state
    #    self.bots = {} # id => bot objects
    #    self.packages = {} # id => package objecs
    #    self.delivered = [] # DEBUG
    #    self.board = None
    #    self.bot = None
    
    #  def read_server_response(self) : pass
    #  def read_packages(self) : pass

  def __getattr__(self, attr):
    """Proxy to singleton"""
    return getattr(self.__single, attr)

  def __setattr__(self, attr, value):
    """Proxy to singleton"""
    setattr(self.__single, attr, value)
    return

  class __real_World:
    """The real World object
       All methods that are called on an instance of World.World() are actually
       implemented here."""

    def __init__(self, descriptor = None):
      self.io = IO.icfpIO(descriptor)
      self.bot = None # us
      self.board = None # state
      self.bots = {} # id => bot objects
      self.py_bots = {} # id(ob) => 1
      self.packages = {} # id => package objecs
      self.delivered = [] # DEBUG

      self.io.write("Player\n")
      line = self.io.readline()
      m = re.match("(\d+)\s+(\d+)", line)
      if (not m):
        raise 'Bad board dimensions!'

      self.board = Board(int(m.group(1)), int(m.group(2)))
      for (i) in range(1,1+int(m.group(2))):
        line = self.io.readline()
        self.board.set_row(i, line)

      self.board.find_bases()
      line = self.io.readline()
      m = re.match("(\d+)\s+(\d+)\s+(\d+)", line)
      if (not m):
        raise 'Bad bot definition'

      self.bot = Bot(int(m.group(1)), int(m.group(2)), int(m.group(3)))
      self.bots[self.bot.id] = self.bot # insert ourselves
      self.py_bots[id(self.bot)] = 1

      self.read_server_response() # read and apply
      self.read_packages() # read and apply
      return

    def read_packages(self):
      """Read packages at the current location, if any"""
      line = self.io.readline()
      if (line == None):
        return None
      m = package_re.match(line)
      if (m):
        if (not self.board.bases.has_key(self.bot.curr_loc)):
          # just a package on the ground, but call it a base
          self.board.bases[self.bot.curr_loc] = 1
      while (m):
        pkg_id = int(m.group(1))
        x_pos = int(m.group(2))
        y_pos = int(m.group(3))
        weight = int(m.group(4))
        line = m.group(5)

        pob = None
        if (self.packages.has_key(pkg_id)):
          pob = self.packages[pkg_id]
        else:
          pob = Package(pkg_id)
          self.packages[pkg_id] = pob
        pob.set_weight(weight)
        pob.set_destination(x_pos, y_pos)
        self.board.extra(pob, self.bot.x_pos, self.bot.y_pos)

        m = package_re.match(line)
      # end while(m)

      return 1
    
    def apply_commands(self, cmds):
      """update state from the server commands"""
      seen_bots = {self.bot.id: 1}
      for (cmd) in cmds:
        if (not self.bots.has_key(cmd[0])):
          newbot = Bot(cmd[0], self.bot.carrymax, self.bot.startmoney) # assume he gets our stats
          self.bots[cmd[0]] = newbot
          self.py_bots[id(newbot)] = 1
        b = self.bots[cmd[0]]
        seen_bots[b.id] = 1 # to see who isn't dead
        # endif (initial)
        if (cmd[1] in ['N','S','E','W']):
          b.move(cmd[1])
          self.board.extra(b, b.x_pos, b.y_pos)
        elif (cmd[1] == 'X'):
          b.set_x(int(cmd[2]))
          # if this was the second, place him
          if (b.x_pos and b.y_pos):
            self.board.extra(b, b.x_pos, b.y_pos)
        elif (cmd[1] == 'Y'):
          b.set_y(int(cmd[2]))
          # if this was the second, place him
          if (b.x_pos and b.y_pos):
            self.board.extra(b, b.x_pos, b.y_pos)
        # P and D are FOOOBARRRED, fix for list
        # They might be fixed?
        elif (cmd[1] == 'P'):
          for (p_id) in cmd[2:]:
            p_id = int(p_id)
            package = None
            if (self.packages.has_key(p_id)):
              package = self.packages[p_id]
            else: # create a new package
              package = Package(p_id)
              self.packages[p_id] = package
          b.pickup(package)
          loc = (self.bot.x_pos, self.bot.y_pos)
          if (not self.board.rm_extra(package)):
            if (self.board.bases.has_key(loc)):
              del self.board.bases[loc]
            if (self.board.maybe_bases.has_key(loc)):
              del self.board.maybe_bases[loc]
            if (self.board.doubtful_bases.has_key(loc)):
              del self.board.doubtful_bases[loc]
        elif (cmd[1] == 'D'):
          for (p_id) in cmd[2:]:
            p_id = int(p_id)
            package = self.packages[p_id]
            b.drop(package)
            if (b.id == self.bot.id): # we dropped it, so we know stuff about it
              if (b.x_pos != package.x_des or b.y_pos != package.y_des):
                # We might have been pushed more than once,
                # but this should put us in the right area
                self.board.maybe_bases[(b.last_loc)] = 1
              else:
                self.delivered.append(package.id)
                del self.packages[p_id]
            else:
              # someone else probably making a delivery, add it to doubtful
              self.board.doubtful_bases[(b.curr_loc)] = 1
        else:
          raise 'Unkown command from server'
      # for (cmd) in cmds

      # We would like to just compare the count, but
      # it could be one bot died and another joined
      for (bot_id) in self.bots.keys():
        if (not seen_bots.has_key(bot_id)):
          self.remove_bot(bot_id)

      return

    def remove_bot(self, bot_id):
      """Remove the bot from the bots list,
         from the board, and if he was carrying packages
         remove those too"""
      bob = self.bots[bot_id]
      del self.bots[bot_id]
      self.board.rm_extra(bob)
      for (package_id) in bob.packages.keys():
        del self.packages[package_id]

      return
    
    def read_server_response(self):
      """Get a line of update info from the server"""

      line = self.io.readline()
      if (line == None):
        return None
      #print "Server response '%s'" % (line)
      updates = line.split('#')
      updates.pop(0)
      all_cmds = []
      for (update) in updates:
        m = id_re.search(update)
        if (not m):
          raise 'Bad line from server'
        id = int(m.group(1))
        rest = m.group(2)
        if (id == self.bot.id and not rest): # DEBUG
          raise 'We scrwed up'
        cmds = filter(len, re.split('\s*(\w+)\s*',rest))
        while (cmds):
          cmd = cmds.pop(0)
          l = [id, cmd]
          if (cmds and re.match('\d+',cmds[0])):
            l.extend(string.split(cmds.pop(0)))
          all_cmds.append(l)
        # while (cmds)

      self.apply_commands(all_cmds)
      return 1
    
